<'
# Copyright (c) 2012-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Mon Apr  9 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#ifdef __linux__
# if defined(HAVE_VASPRINTF) && defined(HAVE_ASPRINTF)
/* Apparently we need _GNU_SOURCE defined to get access to vasprintf on Linux */
#  ifndef _GNU_SOURCE
#   define _GNU_SOURCE
#  endif
# endif
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "<"$comp">/c/client.h"
#include "<"$comp">_json-common.h"
#include "<"$comp">_typecopy.h"

<'foreach s [$component services] {'>

/* <"[--- Service [$s name] encoding/decoding ---------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_json_scan(
  struct genom_<"$comp">_<"[$s name]">_input *in, const char *json,
  char **endptr)
{
  int s = 0;
  json_skip_whitespace(json);
  if (*(json++) != '{') { s = EINVAL; goto done; }
  json_skip_whitespace(json);
  while(*json != '}') {
    if (*(json++) != '"') { s = EINVAL; goto done; };
<'    foreach p [$s parameters in inout] {'>
    if (!strncmp(json, "<"[$p name]">\"", 1+<"[string length [$p name]]">)) {
      json += 1+<"[string length [$p name]]">;
      json_skip_whitespace(json);
      if (*(json++) != ':') { s = EINVAL; goto done; }
      if ((s = json_scan_<"[[$p type] mangle]">(
             <"[[$p type] pass reference in->[$p name]]">, &json)))
        goto done;
    } else
<'    }'>
    { s = EINVAL; goto done; }

    json_skip_whitespace(json);
    if (*json == ',') {
      json++;
      json_skip_whitespace(json);
    }
  }

done:
  if (endptr) *endptr = (char *)json;
  if (s) {
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
  return genom_ok;
}

genom_event
genom_<"$comp">_client_<"[$s name]">_json_print(char **json,
             const struct genom_<"$comp">_<"[$s name]">_output *out)
{
  int s = 0;
  *json = NULL;

  if ((s = bufcat(json, "{"))) {
    if (*json) free(*json);
    *json = NULL;
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

<'  set sep {}'>
<'  foreach p [$s parameters inout out] {'>
  if ((s = bufcat(json, "<"$sep">\"<"[$p name]">\":"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
  if ((s = json_print_<"[[$p type] mangle]">(json,
         <"[[$p type] pass value out->[$p name]]">))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
<'  set sep ,'>
<'    }'>

  if ((s = bufcat(json, "}"))) {
    if (*json) free(*json);
    *json = NULL;
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  return genom_ok;
}

<'}'>

<'foreach p [$component port out] {'>

/* <"[--- Port [$p name] encoding ---------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_json_print(char **json,
  <"[[$p datatype] argument value data]">)
{
  int s;
  *json = NULL;

  if ((s = bufcat(json, "{"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  if ((s = bufcat(json, "\"<"[$p name]">\":"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }
  if ((s = json_print_<"[[$p datatype] mangle]">(json, data))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  if ((s = bufcat(json, "}"))) {
    if (*json) { free(*json); *json = NULL; }
    genom_syserr_detail d = { s };
    return genom_syserr(&d);
  }

  return genom_ok;
}

<'}'>


/* <"[--- ${comp}_json_error --------------------------------------------]"> */

const char *
genom_<"$comp">_client_json_error(genom_event e, const void *detail)
{
  char *s = NULL;

  if (!e) detail = genom_thrown(&e);

  if (e == genom_syserr_id) {
    if (detail) {
      uint32_t e = ((genom_syserr_detail *)detail)->code;
      if (bufcat(&s, "\"%s\"", strerror(e))) return NULL;
    } else {
      if (bufcat(&s, "\"%s\"", genom_syserr_id)) return NULL;
    }
    return s;
  } else if (e == genom_unkex_id && detail) {
    if (detail) {
      if (bufcat(&s, "{\"ex\":\"%s\",\"detail\":{\"what\":\"%s\"}}",
                 genom_unkex_id, ((genom_unkex_detail *)detail)->what))
        return NULL;
    } else {
      if (bufcat(&s, "{\"ex\":\"%s\"}", genom_unkex_id)) return NULL;
    }
    return s;
  }
<'foreach e [$component throws] {'>
  else if (e == <"[$e cname]">_id) {
<'  if {[llength [$e members]]} {'>
    if (detail) {
      if (bufcat(&s, "{\"ex\":\"%s\",\"detail\":", <"[$e cname]">_id))
        return NULL;
      if (json_print_<"[$e mangle]">(&s, (const <"[$e declarator]"> *)detail))
        return NULL;
      if (bufcat(&s, "}")) return NULL;
      return s;
    }
<'  }'>
    if (bufcat(&s, "{\"ex\":\"%s\"}", <"[$e cname]">_id))
      return NULL;
    return s;
  }
<'}'>

  assert(!"unknown genom exception");
  return NULL;
}




/* --- Exception genom::unkex decoding ------------------------------------- */

static int
json_scan_genom_unkex(genom_unkex_detail *exdetail, const char **json)
{
  const char *e;

  json_skip_whitespace(*json);
  if (*((*json)++) != '{') return EINVAL;
  json_skip_whitespace(*json);
  while(**json != '}') {
    if (*((*json)++) != '"') return EINVAL;

    if (!strncmp(*json, "what\"", sizeof("what"))) {
      (*json) += sizeof("what");
      json_skip_whitespace(*json);
      if (*((*json)++) != ':') return EINVAL;
      json_skip_whitespace(*json);
      if (**json != '"') return EINVAL;
      e = ++(*json);
      while (**json != '"') { if (**json == '\\') (*json)++; (*json)++; }
      if ((size_t)(*json - e) >= sizeof(exdetail->what))
        e = *json - sizeof(exdetail->what);
      strncpy(exdetail->what, e, *json - e);
      exdetail->what[*json - e] = *"";

    } else return EINVAL;

    json_skip_whitespace(*json);
    if (**json == ',') {
      (*json)++;
      json_skip_whitespace(*json);
    }
  }

  (*json)++;
  return 0;
}


/* --- Exception genom::syserr decoding ------------------------------------ */

static int
json_scan_genom_syserr(genom_syserr_detail *exdetail, const char **json)
{
  long long v;
  char *d;

  json_skip_whitespace(*json);
  if (*((*json)++) != '{') return EINVAL;
  json_skip_whitespace(*json);
  while(**json != '}') {
    if (*((*json)++) != '"') return EINVAL;

    if (!strncmp(*json, "code\"", sizeof("code"))) {
      (*json) += sizeof("code");
      json_skip_whitespace(*json);
      if (*((*json)++) != ':') return EINVAL;
      json_skip_whitespace(*json);

      v = strtoll(*json, &d, 0);
      if (d == *json) return EINVAL;
      *json = d;
      exdetail->code = (uint32_t)v;

    } else return EINVAL;

    json_skip_whitespace(*json);
    if (**json == ',') {
      (*json)++;
      json_skip_whitespace(*json);
    }
  }

  (*json)++;
  return 0;
}


/* <"[--- genom_${comp}_decodex -----------------------------------------]"> */

void *
genom_<"$comp">_decodex(const char *json, genom_event *ex)
{
  genom_unkex_detail *unkex;
  const char *e;
  int s;

  json_skip_whitespace(json);
  if (*(json++) != '{') { s = EINVAL; goto done; }
  json_skip_whitespace(json);
  while(*json != '}') {
    if (*(json++) != '"') { s = EINVAL; goto done; };

    /* look for exception name */
    if (!strncmp(json, "ex\"", sizeof("ex"))) {
      json += sizeof("ex");
      json_skip_whitespace(json);
      if (*(json++) != ':') { s = EINVAL; goto done; }
      json_skip_whitespace(json);
      if (*json != '"') { s = EINVAL; goto done; }
      e = ++json;
      while (*json != '"') { if (*json == '\\') json++; json++; }
<'foreach t [$component throw] {'>
      if (sizeof(<"[$t cname]">_id) == json - e + 1 &&
          !strncmp(<"[$t cname]">_id, e, sizeof(<"[$t cname]">_id)-1))
        *ex = <"[$t cname]">_id;
      else
<'}'>
      {
        if ((size_t)(json - e) >= sizeof(unkex->what))
          e = json - sizeof(unkex->what);
        unkex = (genom_unkex_detail *)malloc(sizeof(*unkex));
        if (unkex) {
          strncpy(unkex->what, e, json-e);
          unkex->what[sizeof(unkex->what)-1] = *"";
        }
        *ex = genom_unkex_id;
        return unkex;
      }
      json++;

    /* look for exception detail */
    } else if (!strncmp(json, "detail\"", sizeof("detail"))) {
      json += sizeof("detail");
      json_skip_whitespace(json);
      if (*(json++) != ':') { s = EINVAL; goto done; }

<'foreach t [$component throw] {'>
      if (*ex == <"[$t cname]">_id) {
<'  if {[llength [$t members]]} {'>
        <"[$t declarator *detail]">;
        detail = (<"[$t declarator *]">)malloc(sizeof(*detail));
        if (detail) {
          genom::ids::pinit(*detail);
          s = json_scan_<"[$t mangle]">(detail, &json);
          if (s) goto done;
        }
        return detail;
<'  } else {'>
        s = EINVAL; goto done;
<'  }'>
      }
<'}'>
      if (*ex == genom_syserr_id) {
        genom_syserr_detail *detail;
        detail = (genom_syserr_detail *)malloc(sizeof(*detail));
        if (detail) {
          s = json_scan_genom_syserr(detail, &json);
          if (s) goto done;
        }
        return detail;
      }
      if (*ex == genom_unkex_id) {
        genom_unkex_detail *detail;
        detail = (genom_unkex_detail *)malloc(sizeof(*detail));
        if (detail) {
          s = json_scan_genom_unkex(detail, &json);
          if (s) goto done;
        }
        return detail;
      }

    } else { s = EINVAL; goto done; }

    json_skip_whitespace(json);
    if (*json == ',') {
      json++;
      json_skip_whitespace(json);
    }
  }

  if (*ex == genom_ok) {
    unkex = (genom_unkex_detail *)malloc(sizeof(*unkex));
    if (unkex)
      strcpy(unkex->what, "::genom::ok");
    *ex = genom_unkex_id;
    return unkex;
  }
  return NULL;

done:
  {
    genom_syserr_detail *d;
    d = (genom_syserr_detail *)malloc(sizeof(*d));
    if (d) d->code = s;
    *ex = genom_syserr_id;
    return d;
  }
}
