/*
 * Copyright (c) 2012-2013 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Thu Mar 15 2012
 */
#ifndef H_CCLIENT_PRIVATE
#define H_CCLIENT_PRIVATE

#include <pthread.h>

#include <vector>

#include "boost/shared_ptr.hpp"
#include "ros/node_handle.h"
#include "actionlib/client/action_client.h"

#include "genom3/c/client.h"

#define ROS_CLIENT_FREE		0
#define ROS_CLIENT_WAIT_ACK	1
#define ROS_CLIENT_WAIT_TERM	2
#define ROS_CLIENT_TERM		3

struct ros_options {
  char *instance;
};

/* This hackery is required to store a generic pointer on a template <class T>
 * class ClientGoalHandle and invoke cancel() on it, because the actionlib
 * client goal handle does not derive from a common base class... */
class genom_gh {
  struct iface {
    virtual void cancel() = 0;
    virtual ~iface() {}
  };

  template<class T> struct generic : iface {
    generic(const T &t) : gh(t) {}
    void cancel() { return gh.cancel(); }
   private:
    T gh;
  };

  boost::shared_ptr<iface> ghp;

 public:
  genom_gh() : ghp() {}
  template<class T> genom_gh(const T &gh) : ghp(new generic<T>(gh)) {}

  void cancel() { if (ghp) ghp->cancel(); }
  void reset() { if (ghp) ghp.reset(); }
};

/* stores permanent handles to each services/actions, defined in msglib.h */
struct connections;

/* the client */
struct genom_client_s {
  char *instance;
  genom_event error;
  ros::NodeHandle *node;
  struct connections *connections;
  int events;

  struct request {
    const struct genom_service_info *info;
    genom_gh gh;
    int status, pstatus;
    genom_event error;
    const void *exdetail;
    void *output;
    genom_request_cb sentcb, donecb;
    void *cb_data;
  };
  std::vector<request> active_rqst;
  int next;
  pthread_mutex_t activelk;
};

genom_event	ros_init(int argc, char *argv[]);
void		ros_fini(void);
genom_event	ros_getopt(int argc, char *argv[], struct ros_options *opts);

int		ros_eventfd(void);
const struct genom_service_info *
		ros_service_info(genom_client h, int rqstid);
genom_event	ros_waitevent(void);
genom_event	ros_waitack(genom_client h, int rqstid);
genom_event	ros_waitterm(genom_client h, int rqstid);
int		ros_done(genom_client h, int rqstid);
genom_event	ros_clean(genom_client h, int rqstid);
genom_event	ros_doevents(genom_client h);

genom_event	ros_newrqst(genom_client h, size_t outsize, int &rqstid);
void		ros_evtask_signal(void);

#endif /* H_CCLIENT_PRIVATE */
